/* Module:			winfind.c
*  Description:	Windows 'Finder'
*  Author:			Steven Palmer
*  Date:				Saturday 8th February 1992
*  Updated:			Saturday 22nd February 1992
*
*	Note:				This source code belongs in the public domain. It may be
*						altered, but any changes should acknowledge the original
*						source, which is, of course, Steven Palmer.
*
*  Change Log:
*		22/2/1992	Fixed bug which prohibited viewing drives C:, D:, etc if
*						no drive B: existed.
*		22/2/1992	Added code to handle file association.
*		22/2/1992	Added code to show different disk drive icons depending
*						on whether the drive is fixed, remote or removeable.
*		22/2/1992	Added menu bar.
*		22/2/1992	Added code to select files and/or folders, either in
*						singles or as a group. Selected files can be opened from
*						the Open command on the menu.
*/

#include <windows.h>
#include <dos.h>
#include "winfind.h"

#define	GWW_PARENT				0
#define	GWW_CHILD				2
#define	GWW_FILELIST			4
#define	GWW_FILECOUNT			6
#define	GWW_SCROLLPOS			8
#define	GWW_SCROLLEND			10
#define	GWW_OFFSET				12
#define	GWW_PAGEHEIGHT			14
#define	GWW_WNDEXTRA			16

HANDLE hInst;						/* Instance handle of this application */
HWND hwndMain;						/* Invisible main window */
HICON hFolderIcon;				/* Icon representing a folder */
HICON hFileIcon;					/* Icon representing a file */
HICON hSelFolder;					/* Icon representing a selected folder */
HICON hSelFile;					/* Icon representing a selected file */
HICON hFixedIcon;					/* Fixed disk drive icon */
HICON hRemoteIcon;				/* Remote disk drive icon */
HICON hFloppyIcon;				/* Fixed disk drive icon */
HFONT hTextFont;					/* Icon and filename font */
HCURSOR hHourGlass;				/* Hour glass cursor */
HCURSOR hFixedCursor;			/* Fixed disk drive cursor */
HCURSOR hRemoteCursor;			/* Remote disk drive cursor */
HCURSOR hFloppyCursor;			/* Fixed disk drive cursor */

typedef struct tagAFILE {
	int x1, y1, x2, y2;			/* Coordinates of icon */
	int attrib;						/* File attributes */
	BOOL fSelected;				/* Non-zero if icon or file selected */
	HICON hIcon;					/* Icon associated with file */
	HICON hSelIcon;				/* Selected icon */
	HWND hwnd;						/* Window handle of open folder window */
	char sName[ 13 ];				/* File name */
} AFILE;

char szMainClass[] = "mclass";
char szDiskClass[] = "dclass";
char szDirClass[] = "wclass";

int NEAR PASCAL WinMain( HANDLE, HANDLE, LPSTR, int );
BOOL InitializeInstance( void );
LONG FAR PASCAL MainProc( HWND, unsigned, WORD, LONG );
LONG FAR PASCAL DiskProc( HWND, unsigned, WORD, LONG );
LONG FAR PASCAL DirProc( HWND, unsigned, WORD, LONG );
HWND OpenDirectory( HWND, LPSTR );
BOOL CreateDriveIcon( char );
void BuildFileList( HWND );
void ShowDirectory( HWND, HDC );
void ActivateFile( HWND, int, int );
void SelectFile( HWND, int, int );
void UnselectFile( HWND, AFILE _far * );
void OpenSelectFiles( HWND );
void OpenFileOrFolder( HWND, AFILE _far *, int );
int CountSelectFiles( HWND );

/****************************************************************************
* WinMain
*
* Description:
*  This is the main function.
*
* Input:
*	hInstance = handle to this instance of the application
*	hPrevInstance = handle to previous instance of the application
*	lpszCmdLine = far pointer to copy of command line
*	nCmdShow = flag specifying how application is to be opened
*
* Output:
*	Returns exit code from PostQuitMessage - usually always zero.
****************************************************************************/

int NEAR PASCAL WinMain(
	HANDLE hInstance,
	HANDLE hPrevInstance,
	LPSTR lpszCmdLine,
	int nCmdShow )
{
	MSG msg;

	/* Don't allow more than one instance of this app.
	 */
	if( hPrevInstance )
		return( FALSE );

	/* Save the current instance handle.
	 */
	hInst = hInstance;

	/* Do initialization
	 */
	if( !InitializeInstance() )
		return( FALSE );

	/* Get and dispatch messages.
	 */
	while( GetMessage( &msg, NULL, 0, 0 ) )
		{
		TranslateMessage( &msg );
		DispatchMessage( &msg );
		}	/* End of while */

	/* We get here after Quit was posted.
	 */
	return( (int)msg.wParam );
}

/****************************************************************************
* InitalizeInstance
*
* Description:
*	Here, we create the window classes, create the main window (which is
*	invisible), and create icons for all the main disk drives.
*
* Input:
*	None
*
* Return Value
*	The return value is TRUE if the application initialized OK, FALSE if
*	an error occurred.
****************************************************************************/

BOOL InitializeInstance( void )
{
	WNDCLASS	wc;
	register char ch;

	/* Register the main class.
	 */
	wc.style				= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= MainProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInst;
	wc.hIcon				= NULL;
	wc.hCursor			= NULL;
	wc.hbrBackground	= COLOR_WINDOW + 1;
	wc.lpszMenuName	= NULL;
	wc.lpszClassName	= szMainClass;
	if( !RegisterClass( &wc ) )
		return( FALSE );

	/* Register the class for the disk drive icons
	 */
	wc.style				= CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc		= DiskProc;
	wc.cbWndExtra		= GWW_WNDEXTRA;
	wc.hIcon				= NULL;
	wc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground	= COLOR_BACKGROUND + 1;
	wc.lpszClassName	= szDiskClass;
	if( !RegisterClass( &wc ) )
		return( FALSE );

	/* Register the class for the open directory window.
	 */
	wc.style				= CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	wc.lpfnWndProc		= DirProc;
	wc.cbWndExtra		= GWW_WNDEXTRA;
	wc.hIcon				= LoadIcon( hInst, "FOLDER" );
	wc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	wc.lpszMenuName	= "WINFIND";
	wc.hbrBackground	= COLOR_WINDOW + 1;
	wc.lpszClassName	= szDirClass;
	if( !RegisterClass( &wc ) )
		return( FALSE );

	/* Get a handle to all icons and cursors.
	 */
	hFolderIcon = LoadIcon( hInst, "FOLDER" );
	hSelFolder = LoadIcon( hInst, "SELFOLDER" );
	hFileIcon = LoadIcon( hInst, "FILE" );
	hSelFile = LoadIcon( hInst, "SELFILE" );
	hFixedIcon = LoadIcon( hInst, "FIXED" );
	hFloppyIcon = LoadIcon( hInst, "FLOPPY" );
	hRemoteIcon = LoadIcon( hInst, "REMOTE" );
	hHourGlass = LoadCursor( NULL, IDC_WAIT );
	hFixedCursor = LoadCursor( hInst, "FIXEDCUR" );
	hFloppyCursor = LoadCursor( hInst, "FLOPPYCUR" );
	hRemoteCursor = LoadCursor( hInst, "REMOTECUR" );

	/* Create a 10-point Helv font which will be used to show the icon
	 * names. Any font can be used - the display will be scaled to fit.
	 */
	hTextFont = CreateFont( 10, 0, 0, 0, 400, FALSE, FALSE,
									FALSE, ANSI_CHARSET, OUT_CHARACTER_PRECIS,
									CLIP_DEFAULT_PRECIS, PROOF_QUALITY,
									VARIABLE_PITCH | FF_SWISS, "Helv" );
	if( !hTextFont )
		return( FALSE );

	/* Create an invisible main window that will be used as the focus of
	 * all top level messages.
	 */
	hwndMain = CreateWindow( szMainClass, NULL, 0L, 0, 0,
		0, 0, NULL, NULL, hInst, NULL );
	if( !hwndMain )
		return( FALSE );

	/* Create icons for the floppy disk drives.
	 */
	CreateDriveIcon( 'A' );
	CreateDriveIcon( 'B' );

	/* Create icons for all the fixed disk drives
	 */
	for( ch = 'C'; CreateDriveIcon( ch ); ++ch );
	return( TRUE );
}

/****************************************************************************
* CreateDriveIcon
*
* Description:
*	Create an icon for the specified drive.
*
* Input:
*	chDrive = disk drive letter.
*
* Return Value
*	Returns TRUE if the drive exists, FALSE otherwise.
****************************************************************************/

BOOL CreateDriveIcon( char chDrive )
{
	if( GetDriveType( chDrive - 'A' ) > 1 )
		{
		char sz[ 8 ];
		HWND hwnd;

		wsprintf( sz, "Drive %c", chDrive );
		hwnd = CreateWindow( szDiskClass, sz, 0L, CW_USEDEFAULT, 0,
			CW_USEDEFAULT, 0, hwndMain, NULL, hInst, NULL );
		if( !hwnd )
			return( FALSE );
		SetWindowWord( hwnd, GWW_PARENT, chDrive );
		SetWindowWord( hwnd, GWW_CHILD, NULL );
		ShowWindow( hwnd, SW_SHOWMINIMIZED );
		return( TRUE );
		}
	return( FALSE );
}

/****************************************************************************
* MainProc
*
* Description:
*	Main window procedure
*
* Input:
*	hwnd = handle to window sending the message
*	message = message being sent
*	wParam = word parameter
*	lParam = long parameter
*
* Return Value
*	Message return value
****************************************************************************/

LONG FAR PASCAL MainProc(
	HWND hwnd,
	unsigned msg,
	WORD wParam,
	LONG lParam )
{
	BOOL fPassToDWP;

	fPassToDWP = FALSE;
	switch( msg )
		{
		/* Can't open the invisible window.
		 */
		case WM_QUERYOPEN:
			return( FALSE );
			break;

		case WM_DESTROY:
			PostQuitMessage( 0 );
			break;

		default:
			fPassToDWP = TRUE;
			break;
		}
	if( fPassToDWP )
		return( DefWindowProc( hwnd, msg, wParam, lParam ) );
	return( 0 );
}

/****************************************************************************
* DiskProc
*
* Description:
*	Procdure for the disk icons.
*
* Input:
*	hwnd = handle to window sending the message
*	message = message being sent
*	wParam = word parameter
*	lParam = long parameter
*
* Return Value
*	Message return value
****************************************************************************/

LONG FAR PASCAL DiskProc(
	HWND hwnd,
	unsigned msg,
	WORD wParam,
	LONG lParam )
{
	BOOL fPassToDWP;

	fPassToDWP = FALSE;
	switch( msg )
		{
		/* Double-click on the icon.
		 * If the top level folder (aka root directory) window is already open,
		 * then bring it to the top. Otherwise create the window.
		 */
		case WM_QUERYOPEN: {
			HWND hwndWin;

			if( ( hwndWin = GetWindowWord( hwnd, GWW_CHILD ) ) != NULL )
				BringWindowToTop( hwndWin );
			else
				{
				char sz[ 3 ];

				wsprintf( sz, "%c:", GetWindowWord( hwnd, GWW_PARENT ) );
				hwndWin = OpenDirectory( hwnd, sz );
				SetWindowWord( hwnd, GWW_CHILD, hwndWin );
				}
			return( FALSE );
			}

		case WM_QUERYDRAGICON: {
			char chDrive;

			chDrive = (char)GetWindowWord( hwnd, GWW_PARENT );
			switch( GetDriveType( chDrive - 'A' ) )
				{
				case DRIVE_REMOVABLE:	return( hFloppyCursor );
				case DRIVE_FIXED:			return( hFixedCursor );
				case DRIVE_REMOTE:		return( hRemoteCursor );
				}
			return( NULL );
			}

		case WM_PAINT: {
			PAINTSTRUCT ps;
			HDC hdc;
			char chDrive;

			hdc = BeginPaint( hwnd, &ps );
			chDrive = (char)GetWindowWord( hwnd, GWW_PARENT );
			switch( GetDriveType( chDrive - 'A' ) )
				{
				case DRIVE_REMOVABLE:
					DrawIcon( hdc, 0, 0, hFloppyIcon );
					break;

				case DRIVE_FIXED:
					DrawIcon( hdc, 0, 0, hFixedIcon );
					break;

				case DRIVE_REMOTE:
					DrawIcon( hdc, 0, 0, hRemoteIcon );
					break;
				}
			EndPaint( hwnd, &ps );
			break;
			}

		case WM_DESTROY:
			PostQuitMessage( 0 );
			break;

		default:
			fPassToDWP = TRUE;
			break;
		}
	if( fPassToDWP )
		return( DefWindowProc( hwnd, msg, wParam, lParam ) );
	return( 0 );
}

/****************************************************************************
* DirProc
*
* Description:
*	Procdure for the directory folders.
*
* Input:
*	hwnd = handle to window sending the message
*	message = message being sent
*	wParam = word parameter
*	lParam = long parameter
*
* Return Value
*	Message return value
****************************************************************************/

LONG FAR PASCAL DirProc(
	HWND hwnd,
	unsigned msg,
	WORD wParam,
	LONG lParam )
{
	BOOL fPassToDWP;

	fPassToDWP = FALSE;
	switch( msg )
		{
		/* Closing a folder window.
		 * Need to watch out here because a parent folder can be closed while
		 * leaving the child window open. IsWindow() will tell us if the
		 * parent window is open or not.
		 */
		case WM_CLOSE: {
			HWND hwndParent;

			hwndParent = GetWindowWord( hwnd, GWW_PARENT );
			if( IsWindow( GetWindowWord( hwndParent, GWW_CHILD ) ) )
				SetWindowWord( hwndParent, GWW_CHILD, NULL );
			else if( IsWindow( hwndParent ) )
				{
				AFILE _far * lpFileList;
				HANDLE hFileList;
				register int c;

				/* Get the file list for the parent directory of this folder,
				 * and restore the icon back to its 'closed' state. Also set
				 * the hwnd parameter in file list to NULL, to indicate that
				 * the folder is closed.
				 */
				hFileList = GetWindowWord( hwndParent, GWW_FILELIST );
				lpFileList = (AFILE _far *)GlobalLock( hFileList );
				c = GetWindowWord( hwnd, GWW_CHILD );
				UnselectFile( hwndParent, &lpFileList[ c ] );
				lpFileList[ c ].hwnd = NULL;
				GlobalUnlock( hFileList );
				}
			DestroyWindow( hwnd );
			break;
			}

		case WM_INITMENU:
			EnableMenuItem( (HMENU)wParam, IDM_OPEN,
				CountSelectFiles( hwnd ) ? MF_ENABLED : MF_GRAYED );
			break;

		case WM_COMMAND:
			switch( wParam )
				{
				case IDM_OPEN:
					OpenSelectFiles( hwnd );
					break;

				case IDM_EXIT:
					PostQuitMessage( 0 );
					break;
				}
			break;

		case WM_LBUTTONDOWN:
			SelectFile( hwnd, LOWORD( lParam ), HIWORD( lParam ) );
			break;

		case WM_LBUTTONDBLCLK:
			ActivateFile( hwnd, LOWORD( lParam ), HIWORD( lParam ) );
			break;

		case WM_VSCROLL: {
			int wOffset;
			int wPos;

			wOffset = 0;
			wPos = (int)GetWindowWord( hwnd, GWW_SCROLLPOS );
			switch( wParam )
				{
				case SB_LINEDOWN:
					wOffset = 1;
					break;

				case SB_PAGEDOWN:
					wOffset = (int)GetWindowWord( hwnd, GWW_PAGEHEIGHT );
					break;

				case SB_THUMBPOSITION:
					wOffset = LOWORD( lParam ) - wPos;
					break;

				case SB_LINEUP:
					wOffset = -1;
					break;

				case SB_PAGEUP:
					wOffset = -GetWindowWord( hwnd, GWW_PAGEHEIGHT );
					break;
				}
			if( wOffset )
				{
				if( wOffset + wPos < 0 )
					wPos = 0;
				else if( wOffset + wPos > (int)GetWindowWord( hwnd, GWW_SCROLLEND ) )
					wPos = (int)GetWindowWord( hwnd, GWW_SCROLLEND );
				else
					wPos += wOffset;
				if( wPos != (int)GetWindowWord( hwnd, GWW_SCROLLPOS ) )
					{
					SetWindowWord( hwnd, GWW_SCROLLPOS, wPos );
					InvalidateRect( hwnd, NULL, TRUE );
					}
				}
			break;
			}

		case WM_PAINT: {
			PAINTSTRUCT ps;
			HDC hdc;

			hdc = BeginPaint( hwnd, &ps );
			ShowDirectory( hwnd, hdc );
			EndPaint( hwnd, &ps );
			break;
			}

		default:
			fPassToDWP = TRUE;
			break;
		}
	if( fPassToDWP )
		return( DefWindowProc( hwnd, msg, wParam, lParam ) );
	return( 0 );
}

/****************************************************************************
* OpenDirectory
*
* Description:
*	This function opens a directory window for all files and subdirectories
*	on the specified path.
*
* Input:
*	hwndParent = handle of parent window (parent directory or icon)
*	lpszPath = pathname, which must NOT be terminated by backslash (\)
*
* Return Value
*	Returns the handle of the folder window, or NULL if the window could
*	not be created.
****************************************************************************/

HWND OpenDirectory( HWND hwndParent, LPSTR lpszPath )
{
	HWND hwnd;

	hwnd = CreateWindow( szDirClass, lpszPath, WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, NULL, NULL, hInst, NULL );
	if( hwnd )
		{
		SetWindowWord( hwnd, GWW_PARENT, hwndParent );
		SetWindowWord( hwnd, GWW_CHILD, NULL );
		BuildFileList( hwnd );
		ShowWindow( hwnd, SW_SHOW );
		}
	return( hwnd );
}

/****************************************************************************
* BuildFileList
*
* Description:
*	This function constructs an array of all files and subdirectories in the
*	directory viewed by the specified directory window.
*
* Input:
*	hwnd = handle of directory window
*
* Return Value
*	None
****************************************************************************/

void BuildFileList( HWND hwnd )
{
	char sz[ 100 ];
	HANDLE hFileList;
	AFILE _far * lpFileList;
	struct find_t ft;
	HCURSOR hCursorT;
	int cFiles;
	int r;

	/* Compute the wildcard search string from the window title name.
	 */
	GetWindowText( hwnd, sz, sizeof( sz ) - 5 );
	lstrcat( sz, "\\*.*" );

	/* Change to the hour-glass cursor.
	 * The following code could take time to execute...
	 */
	hCursorT = SetCursor( hHourGlass );

	/* Compute the total number of files in the current directory, including
	 * subdirectories. Ignore "." and "..".
	 */
	r = _dos_findfirst( sz, _A_SUBDIR, &ft );
	cFiles = 0;
	while( !r )
		{
		if( lstrcmp( ft.name, "." ) && lstrcmp( ft.name, ".." ) )
			++cFiles;
		r = _dos_findnext( &ft );
		}

	/* Allocate a global memory block to hold information about all files
	 * in the directory.
	 * Lock it to get lpFileList, a FAR pointer.
	 */
	hFileList = GlobalAlloc( GHND, cFiles * sizeof( AFILE ) );
	lpFileList = (AFILE _far *)GlobalLock( hFileList );

	/* Gather all the files, except for the two special cases "." and "..",
	 * into the array.
	 */
	r = _dos_findfirst( sz, _A_SUBDIR, &ft );
	while( !r )
		{
		if( lstrcmp( ft.name, "." ) && lstrcmp( ft.name, ".." ) )
			{
			lstrcpy( lpFileList->sName, ft.name );
			lpFileList->attrib = ft.attrib;
			if( lpFileList->attrib & _A_SUBDIR )
				{
				lpFileList->hIcon = hFolderIcon;
				lpFileList->hSelIcon = hSelFolder;
				}
			else
				{
				lpFileList->hIcon = hFileIcon;
				lpFileList->hSelIcon = hSelFile;
				}
	/* If the parent window of a folder was closed, then re-opened,
	 * need to re-establish the lpFileList->hwnd link.
	 */
			GetWindowText( hwnd, sz, sizeof( sz ) - 5 );
			lstrcat( sz, "\\" );
			lstrcat( sz, ft.name );
			lpFileList->hwnd = FindWindow( szDirClass, sz );
			lpFileList->fSelected = FALSE;
			++lpFileList;
			}
		r = _dos_findnext( &ft );
		}
	GlobalUnlock( hFileList );

	/* Restore the normal cursor.
	 */
	SetCursor( hCursorT );

	/* Initalise the window word fields.
	 */
	SetWindowWord( hwnd, GWW_FILELIST, hFileList );
	SetWindowWord( hwnd, GWW_FILECOUNT, cFiles );
	SetWindowWord( hwnd, GWW_SCROLLPOS, 0 );
	SetWindowWord( hwnd, GWW_OFFSET, 0 );
}

/****************************************************************************
* ShowDirectory
*
* Description:
*	This function fills out a directory window with all files and
*	subdirectories in the directory specified for the window.
*
* Input:
*	hwnd = handle of directory window
*	hdc = directory window display context
*
* Return Value
*	None
****************************************************************************/

void ShowDirectory( HWND hwnd, HDC hdc )
{
	AFILE _far * lpFileList;
	long lExtent;
	register int x, y;
	HANDLE hFileList;
	int cFiles;
	int cxIcon, cyIcon;
	RECT rect;
	int cView, nOff;
	int dxLine, dyLine;
	int dxPage, dyPage;
	int c;

	/* Initalise. Get the client area, initalise the (x,y) of the first icon
	 * and get the pointer to the start of the file list array.
	 */
	GetClientRect( hwnd, &rect );
	x = 5;
	y = 5;
	cFiles = GetWindowWord( hwnd, GWW_FILECOUNT );
	hFileList = GetWindowWord( hwnd, GWW_FILELIST );
	lpFileList = (AFILE _far *)GlobalLock( hFileList );

	/* Compute the best icon spacing based on the width of a full filename.
	 * The optimum would be 'MMMMMMMM.MMM', but this is a bit too extreme.
	 */
	SelectObject( hdc, hTextFont );
	cxIcon = GetSystemMetrics( SM_CXICON );
	cyIcon = GetSystemMetrics( SM_CYICON );
	lExtent = GetTextExtent( hdc, "DDDDDDDD.DDD", 12 );
	dxLine = LOWORD( lExtent );
	dyLine = cyIcon + 10 + HIWORD( lExtent );

	/* Get the current scroll position.
	 */
	nOff = GetWindowWord( hwnd, GWW_SCROLLPOS );

	/* Work out and store the total number of icons in a row. That is,
	 * the total number of icons that are fully visible.
	 */
	dyPage = rect.bottom / dyLine;
	dxPage = rect.right / dxLine;
	SetWindowWord( hwnd, GWW_PAGEHEIGHT, dyPage );

	/* Will we need vertical scroll bars ?
	 * Decide now, so that we can reduce rect.right and avoid the need for
	 * horizontal scroll bars.
	 */
	cView = dxPage * dyPage;
	if( cView < cFiles )
		{
		int nScrlRange;
		int cf;

		rect.right -= GetSystemMetrics( SM_CXVSCROLL );
		dxPage = rect.right / dxLine;
		cView = dxPage * dyPage;
			/* This code sets cf to the next factor of dxPage after cFiles.
			 * ie. cf must divide by dxPage with no remainder.
			 */
		cf = ( ( cFiles + ( dxPage - 1 ) ) / dxPage ) * dxPage;
		nScrlRange = ( cf - cView ) / dxPage;
		SetScrollRange( hwnd, SB_VERT, 0, nScrlRange, FALSE );
		SetScrollPos( hwnd, SB_VERT, nOff, TRUE );
		SetWindowWord( hwnd, GWW_SCROLLEND, nScrlRange );
		}
	else
		{
		SetScrollRange( hwnd, SB_VERT, 0, 0, TRUE );
		SetWindowWord( hwnd, GWW_SCROLLPOS, nOff = 0 );
		SetWindowWord( hwnd, GWW_SCROLLEND, 0 );
		}

	/* Loop for all the files in the array and draw their icons.
	 * - nOff was the row offset of the icon on the top line. If we
	 *   multiply it by dxPage, we get the physical offset in the file
	 *   list array of that icon.
	 */
	nOff *= dxPage;
	lpFileList += nOff;
	SetWindowWord( hwnd, GWW_OFFSET, nOff );
	for( c = nOff; c < cFiles && y < rect.bottom; ++c )
		{
		register int dx;
		register int dy;
		int xT;
		HICON hIcon;

		if( lpFileList->hwnd )
			hIcon = lpFileList->hSelIcon;
		else
			hIcon = lpFileList->hIcon;
		lExtent = GetTextExtent( hdc, lpFileList->sName,
										 lstrlen( lpFileList->sName ) );
		dx = LOWORD( lExtent );
		dy = HIWORD( lExtent );
		if( dx <= dxLine )
			xT = x + ( dxLine - dx ) / 2;
		else
			xT = x;
		if( x + dxLine > rect.right )
			{
			x = 5;
			xT = x + ( dxLine - dx ) / 2;
			y += dyLine;
			}
		lpFileList->x1 = x + ( dxLine - cxIcon ) / 2;
		lpFileList->y1 = y;
		lpFileList->x2 = lpFileList->x1 + cxIcon;
		lpFileList->y2 = lpFileList->y1 + cyIcon;
		DrawIcon( hdc, lpFileList->x1, lpFileList->y1, hIcon );
		TextOut( hdc, xT, y + cyIcon + 5, lpFileList->sName,
					lstrlen( lpFileList->sName ) );
		x += dxLine;
		++lpFileList;
		}
	GlobalUnlock( hFileList );
}

/****************************************************************************
* SelectFile
*
* Description:
*	This function is called when the user clicks in a directory window. It
*	finds the icon under the mouse cursor, and highlights the icon.
*
* Input:
*	hwnd = handle of directory window
*	mx = mouse cursor X coordinate
*	my = mouse cursor Y coordinate
*
* Return Value
*	None
****************************************************************************/

void SelectFile( HWND hwnd, int mx, int my )
{
	AFILE _far * lpFileList;
	HANDLE hFileList;
	int cFiles;
	int nOff;
	int c;

	cFiles = GetWindowWord( hwnd, GWW_FILECOUNT );
	hFileList = GetWindowWord( hwnd, GWW_FILELIST );
	lpFileList = (AFILE _far *)GlobalLock( hFileList );
	nOff = GetWindowWord( hwnd, GWW_OFFSET );
	for( c = 0; c < cFiles; ++c )
		if( mx >= lpFileList[ c ].x1 && mx <= lpFileList[ c ].x2 &&
			 my >= lpFileList[ c ].y1 && my <= lpFileList[ c ].y2 &&
			 c >= nOff )
			{
			HDC hdc;

			hdc = GetDC( hwnd );
			DrawIcon( hdc, lpFileList[ c ].x1, lpFileList[ c ].y1,
				lpFileList[ c ].hSelIcon );
			lpFileList[ c ].fSelected = TRUE;
			ReleaseDC( hwnd, hdc );
			}
		else if( lpFileList[ c ].fSelected )
			{
			if( !( GetKeyState( VK_SHIFT ) & 0x1000 ) )
				UnselectFile( hwnd, &lpFileList[ c ] );
			}
	GlobalUnlock( hFileList );
}

/****************************************************************************
* UnselectFile
*
* Description:
*	Unselects the file in the directory window and restores the icon back
*	to its unselected state.
*
* Input:
*	hwnd = handle of directory window
*	lpFileList = points to entry in the file list
*
* Return Value
*	None
****************************************************************************/

void UnselectFile( HWND hwnd, AFILE _far * lpFileList )
{
	HDC hdc;

	hdc = GetDC( hwnd );
	DrawIcon( hdc, lpFileList->x1, lpFileList->y1, lpFileList->hIcon );
	lpFileList->fSelected = FALSE;
	ReleaseDC( hwnd, hdc );
}

/****************************************************************************
* ActivateFile
*
* Description:
*	This function is called when the user double-clicks in a directory
*	window. It finds the icon under the mouse cursor, and if there is a
*	folder, it opens the folder, otherwise if there is a file, it tries to
*	execute the file. If there is nothing there, it returns.
*
* Input:
*	hwnd = handle of directory window
*	mx = mouse cursor X coordinate
*	my = mouse cursor Y coordinate
*
* Return Value
*	None
****************************************************************************/

void ActivateFile( HWND hwnd, int mx, int my )
{
	AFILE _far * lpFileList;
	HANDLE hFileList;
	int cFiles;
	int nOff;
	int c;

	cFiles = GetWindowWord( hwnd, GWW_FILECOUNT );
	hFileList = GetWindowWord( hwnd, GWW_FILELIST );
	lpFileList = (AFILE _far *)GlobalLock( hFileList );
	nOff = GetWindowWord( hwnd, GWW_OFFSET );
	for( c = nOff; c < cFiles; ++c )
		{
		if( mx >= lpFileList[ c ].x1 && mx <= lpFileList[ c ].x2 &&
			 my >= lpFileList[ c ].y1 && my <= lpFileList[ c ].y2 )
			{
			OpenFileOrFolder( hwnd, &lpFileList[ c ], c );
			break;
			}
		}
	GlobalUnlock( hFileList );
}

/****************************************************************************
* OpenSelectFiles
*
* Description:
*	Opens all selected files.
*
* Input:
*	hwnd = handle of directory window
*
* Return Value
*	None
****************************************************************************/

void OpenSelectFiles( HWND hwnd )
{
	AFILE _far * lpFileList;
	HANDLE hFileList;
	int cFiles;
	int c;

	cFiles = GetWindowWord( hwnd, GWW_FILECOUNT );
	hFileList = GetWindowWord( hwnd, GWW_FILELIST );
	lpFileList = (AFILE _far *)GlobalLock( hFileList );
	for( c = 0; c < cFiles; ++c )
		{
		if( lpFileList[ c ].fSelected )
			OpenFileOrFolder( hwnd, &lpFileList[ c ], c );
		}
	GlobalUnlock( hFileList );
}

/****************************************************************************
* CountSelectFiles
*
* Description:
*	Counts the number of selected files. This is used by the code that
*	enables or disables the Open command on the menu.
*
* Input:
*	hwnd = handle of directory window
*
* Return Value
*	Returns the number of selected files.
****************************************************************************/

int CountSelectFiles( HWND hwnd )
{
	AFILE _far * lpFileList;
	HANDLE hFileList;
	int cSelected;
	int cFiles;
	int c;

	cFiles = GetWindowWord( hwnd, GWW_FILECOUNT );
	hFileList = GetWindowWord( hwnd, GWW_FILELIST );
	lpFileList = (AFILE _far *)GlobalLock( hFileList );
	cSelected = 0;
	for( c = 0; c < cFiles; ++c )
		{
		if( lpFileList[ c ].fSelected )
			++cSelected;
		}
	GlobalUnlock( hFileList );
	return( cSelected );
}

/****************************************************************************
* OpenFileOrFolder
*
* Description:
*	Opens the specified file or folder. If it is a file, then if the file
*	has an association, the file is loaded and run as a command line of the
*	program to which it is associated. Unassociated files which are not
*	executable programs produce an error.
*
*	If it is a folder, then a new window is opened for the files in the
*	folder.
*
* Input:
*	hwnd = handle of directory window
*	lpFileList = points to the entry in the file list
*	c = the index of the entry in the file list
*
* Return Value
*	None
****************************************************************************/

void OpenFileOrFolder( HWND hwnd, AFILE _far * lpFileList, int c )
{
	if( lpFileList->attrib & _A_SUBDIR )
		if( lpFileList->hwnd )
			BringWindowToTop( lpFileList->hwnd );
		else
			{
			char sz[ 100 ];

			GetWindowText( hwnd, sz, sizeof( sz ) - 5 );
			lstrcat( sz, "\\" );
			lstrcat( sz, lpFileList->sName );
			lpFileList->hwnd = OpenDirectory( hwnd, sz );
			SetWindowWord( lpFileList->hwnd, GWW_PARENT, hwnd );
			SetWindowWord( lpFileList->hwnd, GWW_CHILD, c );
			}
	else
		{
		char sz[ 100 ];
		char * psz;
		HCURSOR hCursorT;
		LPSTR lpExt;

		psz = sz;
		UnselectFile( hwnd, lpFileList );
		for( lpExt = lpFileList->sName;
			  *lpExt && *lpExt != '.';
			  ++lpExt );
		if( *lpExt++ == '.' )
			{
			char szT[ 60 ];
			char * pszT;

			if( GetProfileString( "Extensions", lpExt, "",
										 szT, sizeof( szT ) ) )
				{
				pszT = szT;
				while( *pszT && *pszT != '^' )
					*psz++ = *pszT++;
				}
			else if( lstrcmpi( lpExt, "com" ) &&
						lstrcmpi( lpExt, "bat" ) &&
						lstrcmpi( lpExt, "exe" ) )
				{
				MessageBox( hwnd, "No association exists for this file",
					"WinFind", MB_OK|MB_ICONEXCLAMATION );
				return;
				}
			}
		GetWindowText( hwnd, psz, sizeof( sz ) - 5 );
		lstrcat( sz, "\\" );
		lstrcat( sz, lpFileList->sName );
		hCursorT = SetCursor( hHourGlass );
		if( WinExec( sz, SW_SHOW ) <= 32 )
			MessageBox( hwnd, "Cannot run application", "WinFind",
				MB_OK|MB_ICONEXCLAMATION );
		SetCursor( hCursorT );
		}
}
